/***************************************************

  bigrev.c : ビッグリバーシ思考ルーチン

             Copyright (C) 2005,2014 Makoto Hiroi

コンパイル
gcc -O2 -shared -fPIC -obigrev.so bigrev.c

盤面

  0 | 1   2   3   4   5   6   7   8   9   10
 ---+----------------------------------------
 11 | 12  13  14  15  16  17  18  19  20  21
 22 | 23  24  25  26  27  28  29  30  31  32
 33 | 34  35  36  37  38  39  40  41  42  43
 44 | 45  46  47  48  49  50  51  52  53  54
 55 | 56  57  58  59  60  61  62  63  64  65
 66 | 67  68  69  70  71  72  73  74  75  76
 77 | 78  79  80  81  82  83  84  85  86  87
 88 | 89  90  91  92  93  94  95  96  97  98
 99 |100 101 102 103 104 105 106 107 108 109
110 |111 112 113 114 115 116 117 118 119 120
----+----------------------------------------
121 |122 123 124 125 126 127 128 129 130 131 132

****************************************************/
#include <tcl/tcl.h>
#include <string.h>
#include "bigrev.h"

/* 関数宣言（探索の高速化） */
static int search_white( int, int, int );
static int finish_white( int, int );

/* #define DEBUG 1 */

/* αβ値の限界値  */
#define MAX_LIMIT (MAX_VALUE + 100)
#define MIN_LIMIT (MIN_VALUE - 100)

/* 石を置けるか */
#define PUT_BLACK 1
#define PUT_WHITE 2

/* 盤面は大域変数で表す */
char board[B_SIZE];
short black;
short white;

static int  postion[SIZE];             /* 探索領域をセットする */
static int  pos_count;

static int  value_table[SIZE];         /* 評価値をセットする */
static int  depth;                     /* 探索手数 */

static int    w1;      /* 位置数の重み w2 を削除  */
static int    w2;      /* 石数の評価のため w2 を復活させる */

/* 8 方向 */
static int vector[8] = {
   1,   /* 右 */
  -1,   /* 左 */
  -11,   /* 上 */
   11,   /* 下 */
   10,   /* 右上 */
  -10,   /* 左下 */  
 -12,   /* 左上 */
  12,   /* 右下 */
 };

/* 初期化データ */
static const char    init_table[B_SIZE] = {
  WALL, WALL, WALL, WALL, WALL,  WALL,  WALL, WALL, WALL, WALL, WALL,
  WALL, FREE, FREE, FREE, FREE,  FREE,  FREE, FREE, FREE, FREE, FREE,
  WALL, FREE, FREE, FREE, FREE,  FREE,  FREE, FREE, FREE, FREE, FREE,
  WALL, FREE, FREE, FREE, FREE,  FREE,  FREE, FREE, FREE, FREE, FREE,
  WALL, FREE, FREE, FREE, FREE,  FREE,  FREE, FREE, FREE, FREE, FREE,
  WALL, FREE, FREE, FREE, FREE, WHITE, BLACK, FREE, FREE, FREE, FREE,
  WALL, FREE, FREE, FREE, FREE, BLACK, WHITE, FREE, FREE, FREE, FREE,
  WALL, FREE, FREE, FREE, FREE,  FREE,  FREE, FREE, FREE, FREE, FREE,
  WALL, FREE, FREE, FREE, FREE,  FREE,  FREE, FREE, FREE, FREE, FREE,
  WALL, FREE, FREE, FREE, FREE,  FREE,  FREE, FREE, FREE, FREE, FREE,
  WALL, FREE, FREE, FREE, FREE,  FREE,  FREE, FREE, FREE, FREE, FREE,
  WALL, WALL, WALL, WALL, WALL,  WALL,  WALL, WALL, WALL, WALL, WALL, WALL
};

/* 座標と盤面の変換 */
const char zahyou_to_board[SIZE] = {
   12,  13,  14,  15,  16,  17,  18,  19,  20,  21,
   23,  24,  25,  26,  27,  28,  29,  30,  31,  32,
   34,  35,  36,  37,  38,  39,  40,  41,  42,  43,
   45,  46,  47,  48,  49,  50,  51,  52,  53,  54,
   56,  57,  58,  59,  60,  61,  62,  63,  64,  65,
   67,  68,  69,  70,  71,  72,  73,  74,  75,  76,
   78,  79,  80,  81,  82,  83,  84,  85,  86,  87,
   89,  90,  91,  92,  93,  94,  95,  96,  97,  98,
  100, 101, 102, 103, 104, 105, 106, 107, 108, 109,
  111, 112, 113, 114, 115, 116, 117, 118, 119, 120,
 };

const char board_to_zahyou[B_SIZE] = {
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
  -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9,
  -1, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
  -1, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
  -1, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
  -1, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
  -1, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59,
  -1, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69,
  -1, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
  -1, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89,
  -1, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
};

/* データの初期化 */
static void init_data( void )
{
  int    i;
  const char  *ptr = init_table;
  for( i = 0; i < B_SIZE; i++, ptr++ ){
    board[i] = *ptr;
  }
  white = 2;
  black = 2;
}

/* 駒を置ける場所をセットする */
static int set_piece_postion( void )
{
  int    i, c;
  for( c = 0, i = B_START; i < B_END; i++ ){
    if( board[i] == FREE ){
      postion[c] = i;
      value_table[c++] = NO_VALUE;
    }
  }
  value_table[c] = NO_VALUE;
  return pos_count = c;
}

/*
 * 駒を裏返すことができるか
 *
 * 高速化のため黒白同時に調べる
 *
 * 返り値 0 bit 黒  1 : 置ける, 0 : 置けない
 *        1 bit 白  1 : 置ける, 0 : 置けない
 *
 */
static int check_reverse( int num )
{
  int    i, result = 0;
  /* 8 方向をチェックする */
  for( i = 0; i < 8; i++ ){
    int  wc = 0, bc = 0, wf = FALSE, bf = FALSE;
    int  pos = num, dp = vector[i];
    char content;
    while( (content = board[ pos += dp ]) != WALL ){
      if( content == FREE ){
	break;            /* 空きエリア */
      } else if( content == BLACK ){
	/* 黒 */
	bf = TRUE;
	if( !wf ) wc++;
      } else {
	/* 白 */
	wf = TRUE;
	if( !bf ) bc++;
      }
    }
    if( bf && bc ) result |= PUT_BLACK;
    if( wf && wc ) result |= PUT_WHITE;
    if( result == (PUT_BLACK | PUT_WHITE) ) break;
  }
  return result;
}

/* 駒を反転する */
static int reverse_piece( int num, int piece, short *buffer )
{
  int   i, count = 0;
  if( board[num] != FREE ) return 0;      /* 石を置くことはできないよ */
  /* 8 方向をチェックする */
  for( i = 0; i < 8; i++ ){
    int  c = count, flag = FALSE;
    int  pos = num, dp = vector[i];
    char content;
    while( (content = board[ (pos += dp) ]) != WALL ){
      if( content == FREE ){
        break;        /* 空きエリア */
      } else if( content == piece ){
        flag = TRUE; break;  /* 同種の駒を見つけたよ */
      } else {
        buffer[c++] = pos;  /* 異種の駒 */
      }
    }

    if( flag && (c > count) ){
      count = c;        /* 裏返しができたよ */
    }

  }
  if( count ){                   /* 駒を置くことができるよ */
    board[num] = piece;
    for( i = 0; i < count; i++ ){
      short n = buffer[i];
      board[n] = piece;
    }
    if( piece == BLACK ){
      black += count + 1;
      white -= count;
    } else {
      white += count + 1;
      black -= count;
    }
  }
  return count;
}

/* 思考ルーチン専用 */
static int reverse_black( int num, short *buffer )
{
  int  i, count = 0;
  /* 8 方向をチェックする */
  for( i = 0; i < 8; i++ ){
    int  c = count, flag = FALSE;
    int  pos = num, dp = vector[i];
    char content;
    while( (content = board[ (pos += dp) ]) != WALL ){
      if( content == FREE ){
        break;        /* 空きエリア */
      } else if( content == BLACK ){
        flag = TRUE; break;  /* 同種の駒を見つけたよ */
      } else {
        buffer[c++] = pos;  /* 異種の駒 */
      }
    }

    if( flag && (c > count) ){
      count = c;        /* 裏返しができたよ */
    }

  }
  buffer[count] = -1;   /* 終端セット */
  if( count ){          /* 駒を置くことができるよ */
    short *ptr = buffer;
    do {
      board[*ptr++] = BLACK;
    } while( *ptr != -1 );
    board[num] = BLACK;
    black += count + 1;
    white -= count;
  }
  return count;
}

/* 思考ルーチン専用 */
static int reverse_white( int num, short *buffer )
{
  int    i, count = 0;
  /* 8 方向をチェックする */
  for( i = 0; i < 8; i++ ){
    int  c = count, flag = FALSE;
    int  pos = num, dp = vector[i];
    char content;
    while( (content = board[ (pos += dp) ]) != WALL ){
      if( content == FREE ){
        break;        /* 空きエリア */
      } else if( content == WHITE ){
        flag = TRUE; break;  /* 同種の駒を見つけたよ */
      } else {
        buffer[c++] = pos;                 /* 異種の駒 */
      }
    }
    if( flag && (c > count) ){
      count = c;        /* 裏返しができたよ */
    }
  }
  buffer[count] = -1;   /* 終端セット */
  if( count ){          /* 駒を置くことができるよ */
    short *ptr = buffer;
    do {
      board[*ptr++] = WHITE;
    } while( *ptr != -1 );
    board[num] = WHITE;
    white += count + 1;
    black -= count;
  }
  return count;
}

/* 石を置くことが出来るか */
static void put_piece_map( char *map )
{
  /* データ */
  static short scan_table[8][16] = { 
    12,  23,  34,  45,  56,  67,  78,  89, 100, 111, -1, -1,-1, -1,-1, -1,  /* 左から右へ */
    12,  13,  14,  15,  16,  17,  18,  19,  20,  21, -1, -1,-1, -1,-1, -1,  /* 上から下へ */
    21,  32,  43,  54,  65,  76,  87,  98, 109, 120, -1, -1,-1, -1,-1, -1,  /* 右から左へ */
   111, 112, 113, 114, 115, 116, 117, 118, 119, 120, -1, -1,-1, -1,-1, -1,  /* 下から上へ */

    14,  15,  16,  17,  18,  19,  20,  21,  32, 43, 54, 65, 76, 87, 98, -1, /* 上から左下へ */
    19,  18,  17,  16,  15,  14,  13,  12,  23, 34, 45, 56, 67, 78, 89, -1, /* 上から右下へ */
   113, 114, 115, 116, 117, 118, 119, 120, 109, 98, 87, 76, 65, 54, 43, -1, /* 下から左上へ */
   118, 117, 116, 115, 114, 113, 112, 111, 100, 89, 78, 67, 56, 45, 34, -1, /* 下から右上へ */
  };
  static short scan_vector[8] = { 1, 11, -1, -11, 10, 12, -12, -10 };

  int i;
  memset( map, 0, B_SIZE );    /* マップの初期化 */
  for( i = 0; i < 8; i++ ){
    short dp = scan_vector[i];
    short j, pos;
    for( j = 0; (pos = scan_table[i][j]) != -1; j++ ){
      short space_postion = -1;
      char  piece = FREE;
      char  content;
      for( ; (content = board[ pos ]) != WALL; pos += dp ){
	if( content == FREE ){
	  /* 空白位置をセット */
	  space_postion = pos;
	  piece = FREE;
	} else if( space_postion > 0 ){
	  if( piece == FREE ){
	    /* 空白の隣の位置に石がある */
	    piece = content;
	  } else if( piece != content ){
	    /* 今までと違う石があるので空き位置に石を置ける */
	    map[space_postion] |= (piece == WHITE ? PUT_BLACK : PUT_WHITE);
	    space_postion = -1;
	    piece = FREE;
	  }
	}
      }
    }
  }
}

/* 
 * 評価関数（黒有利が正となる）
 *
 */

/*
 * 重みテーブル
 *
 */
static const int weight_table[B_SIZE] = {
  0,   0,  0,  0,  0,  0,  0,  0,   0,  0,  0,
  0, 1000, 20, 20, 20, 20, 20, 20, 20, 20, 1000,
  0,  20,  0,  1,  1,  1,  1,  1,  1,  0,  20,
  0,  20,  1,  1,  1,  1,  1,  1,  1,  1,  20,
  0,  20,  1,  1,  1,  1,  1,  1,  1,  1,  20,
  0,  20,  1,  1,  1,  1,  1,  1,  1,  1,  20,
  0,  20,  1,  1,  1,  1,  1,  1,  1,  1,  20,
  0,  20,  1,  1,  1,  1,  1,  1,  1,  1,  20,
  0,  20,  1,  1,  1,  1,  1,  1,  1,  1,  20,
  0,  20,  0,  1,  1,  1,  1,  1,  1,  0,  20,
  0, 1000, 20, 20, 20, 20, 20, 20, 20, 20, 1000,
  0,   0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,
};

/* 位置テーブル */
#define CORNER 1

static const short postion_table[B_SIZE] = {
  0,  0,      0, 0, 0, 0, 0, 0, 0, 0,   0,
  0, CORNER, 14, 0, 0, 0, 0, 0, 0, 19, CORNER,
  0, 34,     12, 0, 0, 0, 0, 0, 0, 21, 43,
  0,  0,      0, 0, 0, 0, 0, 0, 0,  0,  0,
  0,  0,      0, 0, 0, 0, 0, 0, 0,  0,  0,
  0,  0,      0, 0, 0, 0, 0, 0, 0,  0,  0,
  0,  0,      0, 0, 0, 0, 0, 0, 0,  0,  0,
  0,  0,      0, 0, 0, 0, 0, 0, 0,  0,  0,
  0,  0,      0, 0, 0, 0, 0, 0, 0,  0,  0,
  0, 89,    111, 0, 0, 0, 0, 0, 0,120, 98,
  0, CORNER,113, 0, 0, 0, 0, 0, 0,118, CORNER, 
  0,  0,      0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  
};

static int value_func( char turn )
{
  /* 隅 */
  static const short corner[4] = {12, 21, 111, 120 };
  /* 辺のサーチ */
  static const short edge[4][2] = {
    1, 11,  -1, 11,  1, -11, -1, -11,
  };
  static const short revision[4][3] = {
    24, 13, 23,  31, 20, 32,  101, 100, 112,  108, 109, 119,
  };
  /* 隅と辺Ｃの値を変更 */
  static const int rev_value[4][3] = {
    150, 100, 100, 150, 100, 100, 150, 100, 100, 150, 100, 100, 
  };
  int   bv = 0, wv = 0, bc = 0, wc = 0, i;

  char  map[B_SIZE];
  put_piece_map( map );          /* 石が置けるマップを作成 */

  for( i = B_START; i < B_END; i++ ){
    switch( board[i] ){
      case BLACK:
        bv += weight_table[i]; break;
      case WHITE:
        wv += weight_table[i]; break;
      case FREE:
	/* 星はカウントしない */
	if( weight_table[i] ){
	  if( map[i] & PUT_BLACK ){
	    if( turn == BLACK && postion_table[i] == CORNER ){
	      bv += 250;
	    } else {
	      bc++;
	    }
	  } 
	  if( map[i] & PUT_WHITE ){
	    if( turn == WHITE && postion_table[i] == CORNER ){
	      wv += 250;
	    } else {
	      wc++;
	    }
	  }
	}
	break;
      }
  }
  /* 隅の補正 */
  for( i = 0; i < 4; i++ ){
    int j;
    int k = board[ corner[i] ];
    if( k == FREE ){
      /* 隅が空いているよ */
      for( j = 0; j < 3; j++ ){
	int p = revision[i][j];
	int v = rev_value[i][j];
        switch( board[ p ] ){
	  /* 単独Ｃ打ち、星打ちは相手に有利 */
          case BLACK:
	    if( board[ postion_table[p] ] != BLACK ) wv += v;
            break;
          case WHITE:
            if( board[ postion_table[p] ] != WHITE ) bv += v;
            break;
        }
      }
    } else {
      /* 辺の安定性を考慮（まだ不完全）*/
      for( j = 0; j < 2; j++ ){
	short dp = edge[i][j];
	short p = corner[i];
        int v = 0;
        while( board[ (p += dp) ] == k ){
          v += 90;
        }
	if( v == 810 ) v /= 2;     /* 辺全て同じ石 */
        if( k == WHITE ) wv += v;
        else             bv += v;
      }
    }
  }
  return ((bc - wc) * w1) + (bv - wv) + ((white - black) * w2);
}

/* 結果判定 */
static int result_value( int flag )
{
  int    value;
  int    wc = white;
  int    bc = black;
  if( wc + bc == SIZE || flag == TRUE ){
    if( wc == bc ){
      value = 0;    /* 引き分け */
    } else if( bc > wc ){
      value = MAX_VALUE + bc - wc;  /* 黒勝ち */
    } else {
      value = MIN_VALUE + bc - wc;  /* 白勝ち */
    }
  } else if( wc == 0 ){
    value = MAX_VALUE + bc;
  } else if( bc == 0 ){
    value = MIN_VALUE - wc;
  } else {
    value = NO_VALUE;          /* 決着がついていないよ */
  }
  return value;
}


/*
 * ミニマックスによる探索
 *
 */
static int search_black( int his, int limit, int pass_flag )
{
  int   i, rc = TRUE, point = MIN_LIMIT;
  short save_white = white;
  short save_black = black;
  if( his == depth ){
    return value_func( BLACK );
  }
  for( i = 0; i < pos_count; i++ ){
    int   n = postion[i], value;
    short buffer[SIZE/2];

    /* 駒が置いてあるかチェックする */
    if( board[n] != FREE ) continue;
    /* 駒を置く */
    if( !(rc = reverse_black( n, buffer )) ) continue;

    /* 終了チェック */
    if( (value = result_value( FALSE )) == NO_VALUE ){
      /* 再帰する */
      value = search_white( his + 1, point, FALSE );
    }
    /* 元に戻す */
    {
      int j;
      white = save_white;
      black = save_black;
      for( j = 0; j < rc; j++ ){
	board[ buffer[j] ] = WHITE;
      }
      board[n] = FREE;
    }

    /* ミニマックス心臓部 */
    if( value > point ){
      point = value;
    }
    /* α-β枝刈心臓部 */
    if( point > limit ){
      break;
    }
  }
  if( point == MIN_LIMIT ){
    /* 打つ手がなかった */
    if( pass_flag == TRUE ){
      /* 両者打つ手なし */
      point = result_value( TRUE );    /* 終了 */
    } else {
      /* パスだよ */
      point = search_white( his, MIN_LIMIT, TRUE );
    }
  }
  return point;
}

/* 白番の探索 */
static int search_white( int his, int limit, int pass_flag )
{
  int   i, rc = TRUE, point = MAX_LIMIT;
  short save_white = white;
  short save_black = black;
  if( his == depth ){
    return value_func( WHITE );
  }
  for( i = 0; i < pos_count; i++ ){
    int   n = postion[i], value;
    short buffer[SIZE/2];

    /* 駒が置いてあるかチェックする */
    if( board[n] != FREE ) continue;
    /* 駒を置く */
    if( !(rc = reverse_white( n, buffer )) ) continue;

    /* 終了チェック */
    if( (value = result_value( FALSE )) == NO_VALUE ){
      /* 再帰する */
      value = search_black( his + 1, point, FALSE );
    }
    /* 元に戻す */
    {
      int j;
      white = save_white;
      black = save_black;
      for( j = 0; j < rc; j++ ){
	board[ buffer[j] ] = BLACK;
      }
      board[n] = FREE;
    }

    /* ミニマックス心臓部 */
    if( point > value ){
      point = value;
    }
    /* α-β枝刈心臓部 */
    if( point < limit ){
      break;
    }
  }
  if( point == MAX_LIMIT ){
    /* 打つ手がなかった */
    if( pass_flag == TRUE ){
      /* 両者打つ手無し */
      point = result_value( TRUE );    /* 終了 */
    } else {
      /* パスだよ */
      point = search_black( his, MAX_LIMIT, TRUE );
    }
  }
  return point;
}

/* 読み切り用 */
static int finish_black( int limit, int pass_flag )
{
  int   i, point = MIN_LIMIT;
  short save_white = white;
  short save_black = black;
  for( i = 0; i < pos_count; i++ ){
    int    n = postion[i], value, rc;
    short  buffer[SIZE/2];
    /* 駒が置いてあるかチェックする */
    if( board[n] != FREE ) continue;
    /* 駒を置く */
    if( !(rc = reverse_black( n, buffer )) ) continue;

    /* 終了チェック（石数のみ） */
    if( black + white < SIZE ){
      /* 再帰する */
      value = finish_white( point, FALSE );
    } else {
      value = black - white;
    }
    /* 元に戻す */
    {
      int j;
      black = save_black;
      white = save_white;
      for( j = 0; j < rc; j++ ){
	board[ buffer[j] ] = WHITE;
      }
      board[n] = FREE;
    }
    /* ミニマックス心臓部 */
    if( value > point ){
      point = value;
    }
    /* α-β枝刈心臓部 */
    if( point > limit ){
      break;
    }
  }
  if( point == MIN_LIMIT ){
    /* 打つ手がなかった */
    if( pass_flag == TRUE ){
      /* 両者打つ手なし */
      point = black - white;
    } else {
      point = finish_white( MIN_LIMIT, TRUE );
    }
  }
  return point;
}

/* 白番 */
static int finish_white( int limit, int pass_flag )
{
  int    i, point = MAX_LIMIT;
  short save_white = white;
  short save_black = black;
  for( i = 0; i < pos_count; i++ ){
    int    n = postion[i], value, rc;
    short  buffer[SIZE/2];
    if( board[n] != FREE ) continue;

    if( !(rc = reverse_white( n, buffer )) ) continue;

    /* 終了チェック */
    if( black + white < SIZE ){
      /* 再帰する */
      value = finish_black( point, FALSE );
    } else {
      value = black - white;
    }
    /* 元に戻す */
    {
      int j;
      black = save_black;
      white = save_white;
      for( j = 0; j < rc; j++ ){
	board[ buffer[j] ] = BLACK;
      }
      board[n] = FREE;
    }

    /* ミニマックス心臓部 */
    if( point > value ){
      point = value;
    }
    /* α-β枝刈心臓部 */
    if( point < limit ){
      break;
    }
  }
  if( point == MAX_LIMIT ){
    /* 打つ手がなかった */
    if( pass_flag == TRUE ){
      /* 両者打つ手無し */
      point = black - white;
    } else {
      /* パスだよ */
      point = finish_black( MAX_LIMIT, TRUE );
    }
  }
  return point;
}

/* 単純挿入ソート */
static void sort( int limit )
{
  int    i, j;
  for( i = 1; i < limit; i++ ){
    int    tmp_pos = postion[i];
    int    tmp_value = value_table[i];
    for( j = i - 1; j >= 0 && tmp_value > value_table[j]; j-- ){
      postion[j + 1] = postion[j];
      value_table[j + 1] = value_table[j];
    }
    postion[j + 1] = tmp_pos;
    value_table[j + 1] = tmp_value;
  }
}

static void rsort( int limit )
{
  int    i, j;
  for( i = 1; i < limit; i++ ){
    int    tmp_pos = postion[i];
    int    tmp_value = value_table[i];
    for( j = i - 1; j >= 0 && tmp_value < value_table[j]; j-- ){
      postion[j + 1] = postion[j];
      value_table[j + 1] = value_table[j];
    }
    postion[j + 1] = tmp_pos;
    value_table[j + 1] = tmp_value;
  }
}


/*
 * 読み切り
 *
 */
static int finish_move( int turn )
{
  int    limit = set_piece_postion();
  short  save_white = white;
  short  save_black = black;
  int    i, point = (turn == BLACK ? MIN_LIMIT : MAX_LIMIT);
  for( i = 0; i < limit; i++ ){
    int   rc, value, n = postion[i];
    short buffer[SIZE/2];
    if( board[n] != FREE ) continue;

    /* 駒を置く */
    if( turn == BLACK ){
      rc = reverse_black( n, buffer );
    } else {
      rc = reverse_white( n, buffer );
    }
    if( !rc ) continue;

    /* 終了チェック */
    if( black + white < SIZE ){
      if( turn == BLACK ){
	value = finish_white( point, FALSE );
      } else {
	value = finish_black( point, FALSE );
      }
    } else {
      value = black - white;
    }
    value_table[i] = value;
    /* 元に戻す */
    {
      int  j;
      char piece = (turn == WHITE ? BLACK : WHITE);
      white = save_white;
      black = save_black;
      for( j = 0; j < rc; j++ ){
	board[ buffer[j] ] = piece;
      }
      board[n] = FREE;
    }
    
    if((turn  == BLACK && value > point) ||
       (turn  == WHITE && point > value) ){
      point = value;
    }
  }
  /* NO_VALUE を後ろに集める */
  rsort( limit );
  if( value_table[0] == NO_VALUE ){
    /* パスだよ */
    return PASS;
  }
  /* NO_VALUE を除外する */
  while( value_table[limit - 1] == NO_VALUE ) limit--;
  /* ソートする */
  if( turn == BLACK ) sort( limit ); else rsort( limit );
}

/*
 * 反復深化による探索
 *
 */
static int search_move( int turn, int max_depth )
{
  int   limit = set_piece_postion();
  short save_white = white;
  short save_black = black;

  /* 1 -> 3 -> 5, 2 -> 4 と反復深化させる */  
  depth = ((max_depth & 0x01) ? 1 : 2);
  for( ; depth <= max_depth; depth += 2 ){
    int  i, point = (turn == BLACK ? MIN_LIMIT : MAX_LIMIT);
    for( i = 0; i < limit; i++ ){
      int   value, rc, n = postion[i];
      short buffer[SIZE/2];
      if( board[n] != FREE ) continue;

      /* 駒を置く */
      if( turn == BLACK ){
	rc = reverse_black( n, buffer );
      } else {
	rc = reverse_white( n, buffer );
      }
      if( !rc ) continue;

      /* 終了チェック */
      if( (value = result_value( FALSE )) == NO_VALUE ){
        /* 再帰する */
	if( turn == BLACK ){
	  value = search_white( 1, point, FALSE );
	} else {
	  value = search_black( 1, point, FALSE );
	}
      }
      value_table[i] = value;

      /* 元に戻す */
      {
	int  j;
	char piece = (turn == WHITE ? BLACK : WHITE);
	white = save_white;
	black = save_black;
	for( j = 0; j < rc; j++ ){
	  board[ buffer[j] ] = piece;
	}
	board[n] = FREE;
      }

      if((turn  == BLACK && value > point) ||
         (turn  == WHITE && point > value) ){
        point = value;
      }
    }
    if( depth == 1 || depth == 2 ){
      /* NO_VALUE を後ろに集める */
      rsort( limit );
      if( value_table[0] == NO_VALUE ){
        /* パスだよ */
        return PASS;
      }
      /* NO_VALUE を最初の探索から除外する */
      while( value_table[limit - 1] == NO_VALUE ) limit--;
    }
    /* ソートする */
    if( turn == BLACK ) sort( limit ); else rsort( limit );
  }
}


/* 裏返す石をリストに格納 */
static void set_reverse_list( Tcl_Interp *interp, int n, int p )
{
  static Tcl_Obj *buff[SIZE];
  static short piece_buffer[SIZE];
  Tcl_Obj *result;
  int i , r;
  r = reverse_piece( n, p, piece_buffer );
  buff[0] = Tcl_NewIntObj( board_to_zahyou[n] );
  for( i = 0; i < r; i++ ){
    buff[i + 1] = Tcl_NewIntObj( board_to_zahyou[ piece_buffer[i] ] );
  }
  result = Tcl_NewListObj( r + 1, buff );
  Tcl_SetObjResult( interp, result );
  /* 駒数をセット */
  result = Tcl_NewIntObj( white );
  Tcl_SetVar2Ex(interp, "white_piece", NULL, result, TCL_LEAVE_ERR_MSG);
  result = Tcl_NewIntObj( black );
  Tcl_SetVar2Ex(interp, "black_piece", NULL, result, TCL_LEAVE_ERR_MSG);
}

/* 駒を置くことができる場所を数える */
static int count_postion( int piece )
{
  int	i, count = 0;
  for( i = B_START; i < B_END; i++ ){
    if( board[i] != FREE ) continue;
    if( check_reverse( i ) & (piece + 1) ) count++;
  }
  return count;
}

/*
 * Tcl Command
 *
 * init_game             : ゲームの初期化
 * think $level $turn    : 思考ルーチン
 * reverse $n $piece     : ひっくり返る位置をリストで返す
 * check_reverse $n $p   : 駒を裏返すことができるか
 * count_place $turn     : 駒を置ける場所の数
 *
 */

/*
 * Tcl Command : init_game
 *
 */
static int InitGame(ClientData clientData,
		    Tcl_Interp *interp,
		    int objc,
		    Tcl_Obj *CONST objv[])
{
  if( objc != 1 ){
    Tcl_WrongNumArgs(interp, 1, objv, "Usage : init_game");
    return TCL_ERROR;
  }
  init_data();
  return TCL_OK;
}

/*
 * Tcl Command : count_place $turn
 *
 */
static int CountPlace(ClientData clientData,
		      Tcl_Interp *interp,
		      int objc,
		      Tcl_Obj *CONST objv[])
{
  Tcl_Obj *result;
  int turn, n;
  if( objc != 2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "Usage : count_place $turn");
    return TCL_ERROR;
  }
  if( Tcl_GetIntFromObj(interp, objv[1], &turn ) != TCL_OK ){
    return TCL_ERROR;
  }
  if( turn != 0 && turn != 1 ){
    return TCL_ERROR;
  }
  n = count_postion( turn );
  result = Tcl_NewIntObj( n );
  Tcl_SetObjResult( interp, result );
  return TCL_OK;
}

/*
 * Tcl Command : check_reverse $n $piece
 *
 */
static int CheckReverse(ClientData clientData,
		        Tcl_Interp *interp,
		        int objc,
		        Tcl_Obj *CONST objv[])
{
  Tcl_Obj *result;
  int n, p, r;
  if( objc != 3 ){
    Tcl_WrongNumArgs(interp, 1, objv, "Usage : check_reverse $n $piece");
    return TCL_ERROR;
  }
  if( Tcl_GetIntFromObj(interp, objv[1], &n ) != TCL_OK ){
    return TCL_ERROR;
  }
  if( Tcl_GetIntFromObj(interp, objv[2], &p ) != TCL_OK ){
    return TCL_ERROR;
  }
  if( n < 0 || n >= SIZE || (p != 0 && p != 1) ){
    return TCL_ERROR;
  }
  n = zahyou_to_board[n];
  if( board[n] == FREE && (check_reverse( n ) & (p + 1)) ){
    r = 1;
  } else {
    r = 0;
  }
  result = Tcl_NewIntObj( r );
  Tcl_SetObjResult( interp, result );
  return TCL_OK;
}

/*
 * Tcl Command : reverse $n $piece
 *
 */
static int Reverse(ClientData clientData,
		   Tcl_Interp *interp,
		   int objc,
		   Tcl_Obj *CONST objv[])
{
  static Tcl_Obj *buff[SIZE];
  Tcl_Obj *result;
  int i, n, p, r;
  if( objc != 3 ){
    Tcl_WrongNumArgs(interp, 1, objv, "Usage : reverse $n $piece");
    return TCL_ERROR;
  }
  if( Tcl_GetIntFromObj(interp, objv[1], &n ) != TCL_OK ){
    return TCL_ERROR;
  }
  if( Tcl_GetIntFromObj(interp, objv[2], &p ) != TCL_OK ){
    return TCL_ERROR;
  }
  if( n < 0 || n >= SIZE || (p != 0 && p!= 1) ){
    return TCL_ERROR;
  }
  set_reverse_list( interp, zahyou_to_board[n], p );
  return TCL_OK;
}

/* 指手の決定 */
static void decide_move( Tcl_Interp *interp, int turn, int random )
{
  int move = 0, i = 1, v = value_table[0];
  while( v == value_table[i] ) i++;
  /* 同じ値の手がある */
  if( i > 1 ){
    if( random ){
      /* 乱数で指手を決定する */
      move = rand() % i;
    }
  }
  /* 実際に動かす */
  set_reverse_list( interp, postion[move], turn );
}

/*
 * Tcl Command : think $level $turn $random
 *
 */
static int ThinkMove(ClientData clientData,
		     Tcl_Interp *interp,
		     int objc,
		     Tcl_Obj *CONST objv[])
{
  int num  = white + black;
  int rest = SIZE - num;
  int level, turn, random;
  Tcl_Obj *result;
  if( objc != 4 ){
    Tcl_WrongNumArgs(interp, 1, objv, "Usage : think $level $turn $random");
    return TCL_ERROR;
  }
  if( Tcl_GetIntFromObj(interp, objv[1], &level ) != TCL_OK ){
    return TCL_ERROR;
  }
  if( level < 0 ){
    return TCL_ERROR;
  }
  if( Tcl_GetIntFromObj(interp, objv[2], &turn ) != TCL_OK ){
    return TCL_ERROR;
  }
  if( turn != 0 && turn != 1 ){
    return TCL_ERROR;
  }
  if( Tcl_GetIntFromObj(interp, objv[3], &random ) != TCL_OK ){
    return TCL_ERROR;
  }
  if( random != 0 && random != 1 ){
    return TCL_ERROR;
  }
  /* 先読み手数は level + 1 */
  level++;
  /* 重みの設定 */
  if( num + level > 60 ){
    w1 = 30; w2 = 20;
  } else if( num + level > 30 ){
    w1 = 20; w2 = 20;
  } else {
    w1 = 20; w2 = 10;
  }
  /* 読み切りモードのチェック */
  if( level > 6 ){
    finish_move( turn );
  } else {
    search_move( turn, level );
  }
  /* 評価値を Tcl 変数へセットする */
  result = Tcl_NewIntObj( value_table[0] );
  Tcl_SetVar2Ex(interp, "VALUE", NULL, result, TCL_LEAVE_ERR_MSG);
  /* 指し手の決定 */
  decide_move( interp, turn, random );
  return TCL_OK;
}

/*
 * Tcl コマンドの初期化
 *
 */
int Bigrev_Init( Tcl_Interp* interp )
{
  srand( time( NULL ) ); /* 乱数の初期化 */
  Tcl_CreateObjCommand(interp, "init_game", InitGame, NULL, NULL);
  Tcl_CreateObjCommand(interp, "reverse", Reverse, NULL, NULL);
  Tcl_CreateObjCommand(interp, "check_reverse", CheckReverse, NULL, NULL);
  Tcl_CreateObjCommand(interp, "think", ThinkMove, NULL, NULL);
  Tcl_CreateObjCommand(interp, "count_place", CountPlace, NULL, NULL);
  return Tcl_PkgProvide(interp, "bigrev", "1.0");
}

/* end of file */
